
import random
from snake_wars.commons import Location


class RandomLocation(Location):
    """
    A custom Coordinates class to instantiate automatically random
    coordinates from a given range.
    """

    def __init__(self, max_x: int = None, max_y: int = None):

        # Generate random coordinates given maximum coordinates
        x = random.randint(0, max_x)
        y = random.randint(0, max_y)

        super().__init__(x, y)




import sys
import pygame
from multiprocessing import Process
from PodSixNet.Connection import connection, ConnectionListener

from snake_wars.commons import Direction, Size, Location
from snake_wars.client.entities import Snake, Food
from snake_wars.client import Renderer


class Client(ConnectionListener, Process):

    def __init__(self, screen_width: int = 480, screen_height: int = 480,
                 ip: str = "127.0.0.1", port: int = 5071):

        Process.__init__(self)
        ConnectionListener.__init__(self)

        # Set the screen size
        self.screen_size = Size(screen_width, screen_height)

        # Init the id and grid_size, defined later by the server
        self.id = None
        self.grid_size = None
        self.renderer = None
        self.is_connected = False

        # Init a dict to store all snake.
        self.snakes = {}
        self.foods = {}

        # Connect to the server
        self.Connect(address=(ip, port))
        print(f"[Client] Starting complete > Connected to: {ip}:{port}")

    def run(self):
        self.loop()

    def loop(self):
        """Client loop. Refresh the game from the server."""

        while True:
            connection.Pump()
            self.Pump()

            if self.is_connected:
                self.handle_keys()
                self.renderer.render(self.snakes.values(), self.foods)

    def handle_keys(self):
        """Handle the client key press."""

        for event in pygame.event.get():

            if event.type == pygame.QUIT:
                self.quit()

            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_UP:
                    connection.Send({"action": "turn", "message": Direction.UP.value})

                elif event.key == pygame.K_DOWN:
                    connection.Send({"action": "turn", "message": Direction.DOWN.value})

                elif event.key == pygame.K_LEFT:
                    connection.Send({"action": "turn", "message": Direction.LEFT.value})

                elif event.key == pygame.K_RIGHT:
                    connection.Send({"action": "turn", "message": Direction.RIGHT.value})

    @staticmethod
    def quit():
        """Close the Client."""

        pygame.quit()
        sys.exit()

    # NETWORK related functions
    # -------------------------------------------------------------------------

    def Network_authentication(self, data: dict):
        """
        Function triggered when the server respond to this client connection.
        The server return a randomly generated id and the size of the game's
        grid.

        :param data: The data send by the server.
        """
        message = data['message']

        # Save its own id and the game grid size
        self.id = message['id']
        self.grid_size = Size(
            message['grid_size'][0],
            message['grid_size'][1]
        )

        # Instantiate the renderer and set is_connected to True
        self.renderer = Renderer(self.screen_size, self.grid_size)
        self.is_connected = True

    def Network_add_players(self, data: dict):
        """
        Function triggered when the server send the information that
        a new player joined the game.

        :param data: The data send by the server.
        """
        message = data['message']

        # Create a local copy of the given snakes
        for player in message:

            self.snakes[player['id']] = Snake(
                Location(player['location'][0], player['location'][1])
            )

    def Network_update_positions(self, data: dict):
        """
        Function triggered when the server send the positions of all
        Snakes in the game.

        :param data: The data send by the server.
        """
        message: dict = data['message']

        # Reset and create a local copy of each snake
        self.snakes = {}
        for player in message['players']:

            snake = Snake()
            snake.set_positions(player['positions'])

            self.snakes[player['id']] = snake

        # Reset and create a local copy of the given foods
        self.foods = {}
        for food in message['foods']:
            self.foods[food] = Food()

    def Network_game_over(self, data: dict):
        """
        Function triggered when the server send a game over.
        Close the game client.

        :param data: The data send by the server.
        """

        self.quit()




import pygame
from typing import List, Dict
from snake_wars.client.entities import Snake, Food
from snake_wars.commons import Size


class Renderer:
    """PyGame renderer."""

    def __init__(self, screen_size: Size, grid_size: Size):

        # Store the screen size, game grid size, and math the size of a cell
        self.screen_size = screen_size
        self.grid_size = grid_size
        self.cell_size = self.__math_cell_size()

        # Init a pygame window
        pygame.init()

        # PyGame related values
        self.clock = pygame.time.Clock()
        self.screen = pygame.display.set_mode(screen_size.tuple(), 0, 32)

        self.surface = pygame.Surface(self.screen.get_size()).convert()
        self.font = pygame.font.SysFont("monospace", 16)

    def start(self):
        self.__draw_grid()

    def __math_cell_size(self) -> Size:
        """
        Math and return the size of a cell from the size of the
        window and the size of the grid.
        """

        return Size(
            self.screen_size.width // self.grid_size.width,
            self.screen_size.height // self.grid_size.height
        )

    def render(self, snakes: List[Snake], foods: Dict[tuple, Food]):
        self.__draw_grid()
        self.__draw_foods(foods)
        self.__draw_snakes(snakes)

        # text = self.myfont.render("Score {0}".format(snake.score), 1, (0, 0, 0))
        self.screen.blit(self.surface, (0, 0))
        pygame.display.update()

    def __draw_grid(self):

        for h in range(0, self.grid_size.height):
            for w in range(0, self.grid_size.width):

                rect = pygame.Rect(
                    (w * self.cell_size.width, h * self.cell_size.height),
                    (self.screen_size.width, self.screen_size.height)
                )

                if (w + h) % 2 == 0:
                    pygame.draw.rect(self.surface, (93, 216, 228), rect)
                else:
                    pygame.draw.rect(self.surface, (84, 194, 205), rect)

    def __draw_snakes(self, snakes: List[Snake]):

        for snake in snakes:
            for loc in snake.positions:

                rect = pygame.Rect(
                    loc.x * self.cell_size.width,
                    loc.y * self.cell_size.height,
                    self.cell_size.width,
                    self.cell_size.height
                )

                pygame.draw.rect(self.surface, snake.color, rect)
                pygame.draw.rect(self.surface, (93, 216, 228), rect, 1)

    def __draw_foods(self, foods: Dict[tuple, Food]):

        for location, food in foods.items():

            rect = pygame.Rect(
                location[0] * self.cell_size.width,
                location[1] * self.cell_size.height,
                self.cell_size.width,
                self.cell_size.height
            )

            pygame.draw.rect(self.surface, food.color, rect)
            pygame.draw.rect(self.surface, (93, 216, 228), rect, 1)




from typing import Iterable
from snake_wars.commons import Location


class Snake:

    def __init__(self, start_location: Location = None):

        self.positions = [start_location]
        self.color = (17, 24, 47)

    def set_positions(self, positions: list):
        """
        Update the positions of the Snake from a given raw list of positions.
        :param positions: A list of raw positions.
        """

        # TODO: Create a lambda function
        def generate_positions() -> Iterable:
            for position in positions:
                yield Location(position[0], position[1])

        self.positions = list(generate_positions())





import time
import random
from typing import Dict, Iterable
from multiprocessing import Process
from iteration_utilities import duplicates
from PodSixNet.Server import Server as PodSixServer

from snake_wars.commons import RandomLocation, Size
from snake_wars.server.entities import Food
from snake_wars.server import Player


class Server(PodSixServer, Process):

    channelClass = Player

    def __init__(self, ip: str = "127.0.0.1", port: int = 5071, slots: int = 1,
                 grid_width: int = 20, grid_height: int = 20,
                 food_spawn_rate: float = 0.15):

        Process.__init__(self)
        PodSixServer.__init__(self, localaddr=(ip, port))

        # Server address
        self.ip: str = ip
        self.port: int = port

        # Server state
        self.max_slots: int = slots

        # Players (snakes) and food
        self.players: Dict[int, Player] = {}
        self.foods = {}
        self.occupied_positions = []

        # Game constants
        self.grid_size = Size(grid_width, grid_height)
        self.food_spawn_rate: float = food_spawn_rate

        print(f"[Server] Starting complete > Listening to: {self.ip}:{self.port}")

    def run(self):

        while True:
            self.lobby_loop()
            self.game_loop()

    def lobby_loop(self):
        """Lobby loop: Run while waiting for all players to connect."""

        while True:
            self.Pump()

            if len(self.players) == self.max_slots:
                break

    def game_loop(self):
        """
        Game loop: Refresh and update the game state. Send the
        result to all clients.
        """

        while True:
            time.sleep(0.5)
            self.Pump()

            # Refresh player states
            self.__move_all()
            self.__eat_all()
            self.__are_death()
            self.__random_spawn_food()

            # Send updates to players
            self.update_positions()
            self.Pump()

            # Close game if all players disconnected
            if len(self.players) == 0:
                break

    def __move_all(self):
        self.occupied_positions = []

        # Move all players, get their positions
        for player in self.players.values():
            self.occupied_positions += player.move()

        # Remove the Snake that overlap another
        for pos in duplicates(self.occupied_positions):
            for player in self.players.values():

                if pos == player.snake.get_head_position().tuple():
                    player.snake.death = True

    def __eat_all(self):
        for player in self.players.values():

            if not player.snake.death:
                if pos := player.snake.is_eating_food(self.foods):

                    del self.foods[pos]

    def __are_death(self):

        players = {**self.players}  # Create a shallow copy
        for player_id, player in players.items():

            if player.snake.death:
                self.disconnect(player)
                del self.players[player_id]

    def __random_spawn_food(self):
        """
        Randomly spawn a food (or not), at a random position.
        """

        # Boolean random based on probability:
        if random.random() < self.food_spawn_rate:

            # Spawn a food at a random location
            location = RandomLocation(self.grid_size.width, self.grid_size.height)
            self.foods[location.tuple()] = Food(location)

    # NETWORK related functions
    # -------------------------------------------------------------------------

    def Connected(self, new_player: Player, address):
        """
        Function triggered when a new player connect to the server:
        Create a new player in the server; Send a message to all players
        about this new player; And send back a response to the newly
        connected player.

        :param new_player: A player channel Object.
        :param address: A tuple containing the (ip, port) the client connected to.
        :type new_player: ClientChannel
        :type address: tuple
        """

        print(f'[Server] ++ New client connected ++ Socket: {address[1]} - Id: {new_player.id}')

        # Create a Snake for this player
        new_player.create_snake()

        # Send the new player to all connected players
        for player in self.players.values():

            player.Send({"action": "add_players", "message": [{
                'id': new_player.id,
                'location': new_player.snake.get_head_position().tuple(),
            }]})

        # Save this new player
        self.players[new_player.id] = new_player

        # Send its id and the game size to the new player
        new_player.Send({"action": "authentication", "message": {
            'id': new_player.id,
            'grid_size': self.grid_size.tuple(),
        }})

        # Send all the players to the new player
        new_player.Send({"action": "add_players", "message": list(self.__get_all_players_data())})

    def __get_all_players_data(self) -> Iterable:
        """Return an iterator of all player's id and spawning location."""

        for player_id, player in self.players.items():

            yield {
                'id': player_id,
                'location': player.snake.get_head_position().tuple()
            }

    def update_positions(self):
        """Send all positions of all Snakes and foods to all players."""

        # Get all player's snake positions
        all_players_positions = list(self.get_all_players_positions())

        # Send these positions to all players
        for player in self.players.values():
            player.Send({"action": "update_positions", "message": {
                'players': all_players_positions,
                'foods': list(self.foods.keys())
            }})

    def get_all_players_positions(self) -> Iterable:
        """Return an Iterable of all player's positions."""

        for player_id, player in self.players.items():

            yield {
                'id': player_id,
                'positions': list(player.snake.get_all_raw_positions())
            }

    @staticmethod
    def disconnect(player: Player):
        """
        Send a Game Over to a death player. The player should
        disconnect once he receive this message.
        """

        player.Send({"action": "game_over"})




class Location:
    """
    A simple 2D-coordinates class. Useful to store a position,
    or the size of a 2D grid.
    """

    def __init__(self, x: int = 0, y: int = 0):
        self.x = x
        self.y = y

    def tuple(self) -> tuple:
        """Return a tuple of coordinates."""

        return self.x, self.y




import random
from typing import Union
from PodSixNet.Channel import Channel

from snake_wars.server.entities import Snake
from snake_wars.commons import RandomLocation, Size


class Player(Channel):
    """Player socket representation on the server."""

    def __init__(self, *args, **kwargs):

        # Create a random id
        self.id = random.randint(0, 10000)
        self.snake: Union[Snake, None] = None

        super().__init__(*args, **kwargs)

    def create_snake(self):
        """Create a Snake at a random location in the game grid."""

        # Create a random spawn location for the Snake
        grid_size: Size = self._server.grid_size
        location = RandomLocation(grid_size.width, grid_size.height)

        # Instantiate a new Snake
        self.snake = Snake(location, grid_size)

    def move(self):
        """Move its Snake and return its positions."""

        # Move the snake
        self.snake.move()

        # If the snake is still alive, return its positions
        if not self.snake.death:
            return list(self.snake.get_all_raw_positions())

        # If not, return an empty list
        return []

    # NETWORK related functions
    # -------------------------------------------------------------------------

    def Network(self, data):
        pass

    def Network_turn(self, data):
        """Handle client's input (turn left, right, top, bottom)."""

        self.snake.turn(data["message"])




from enum import Flag


class Direction(Flag):
    """Enumerate the movement's direction of a snake."""

    # Direction = (x, y)
    UP = (0, -1)
    DOWN = (0, 1)
    LEFT = (-1, 0)
    RIGHT = (1, 0)

    @staticmethod
    def list():
        """Return a list of all values."""

        return [d.value for d in Direction]




from snake_wars.commons.direction import Direction
from snake_wars.commons.location import Location
from snake_wars.commons.random_location import RandomLocation
from snake_wars.commons.size import Size




from snake_wars.commons import Location


class Size(Location):

    def __init__(self, width: int = 0, height: int = 0):
        super().__init__(width, height)

    @property
    def width(self):
        return self.x

    @property
    def height(self):
        return self.y




import random
from typing import Iterable, List
from snake_wars.commons import Direction, Location, Size


class Snake:

    def __init__(self, start_location: Location, grid_size: Size):

        self.grid_size = grid_size

        # Set position, direction and color of the snake
        self.positions: List[Location] = [start_location]
        self.direction = random.choice(Direction.list())
        self.color = (17, 24, 47)

        # Init the length and score values
        self.length = 1
        self.score = 0
        self.death = False

    def get_head_position(self) -> Location:
        """Return the location of the Snake's head."""

        return self.positions[0]

    def get_all_raw_positions(self) -> Iterable[tuple]:
        """Return an Iterable of all positions as tuples."""

        for position in self.positions:
            yield position.tuple()

    def turn(self, direction: tuple):

        # Avoid the snake to turn on 180 degrees
        if self.length > 1 and (direction[0] * -1, direction[1] * -1) == self.direction:
            return

        # Else, set the new direction
        self.direction = direction

    def move(self):

        current_loc = self.get_head_position()
        x, y = self.direction

        # Math new head position
        # Modulo: When the snake reach a border, it appear on the other side.
        new_loc = Location(
            (current_loc.x + x) % self.grid_size.width,
            (current_loc.y + y) % self.grid_size.height
        )

        # If the snake eat itself, die
        if len(self.positions) > 2 and \
                new_loc.tuple() in list(self.get_all_raw_positions())[2:]:

            self.death = True

        # Else, move once cell ahead
        else:
            self.positions.insert(0, new_loc)

            if len(self.positions) > self.length:
                self.positions.pop()

    def is_eating_food(self, foods: dict):
        current_pos = self.get_head_position().tuple()

        for pos in foods.keys():
            if current_pos == pos:
                self.length += 1
                self.score += 1

                return pos
